"""
Tools to read and write audio files from many places in many different formats.
"""
from hear.stores import *
from hear.tools import AudioSegments, wf_func_to_wfsr_func
from hum.sound import Sound



"""
Accessing segments of audio.
"""
from typing import Callable, Any, Tuple
from functools import lru_cache, wraps, partial
from dataclasses import dataclass

import soundfile as sf

from hear.util import Number, Waveform

AudioSource, BT, TT, Sref, AudioSegmentKey = Any, Any, Any, Any, Any
NormalizedSegmentKey = Tuple[Sref, BT, TT]
SegmentKeyNormalizer = Callable[[AudioSegmentKey], NormalizedSegmentKey]

dflt_key_to_src_bt_tt: SegmentKeyNormalizer


def dflt_key_to_src_bt_tt(k: AudioSegmentKey) -> NormalizedSegmentKey:
    """extract (src, bt, tt) from key (for use in __getitem__)"""
    if isinstance(k, tuple):
        if len(k) == 3:  # assume k is (wf_key, bt, tt)
            return k
        elif len(k) == 2:  # assume k is (wf_key, slice(bt, tt))
            wf_key, _slice = k
            return wf_key, _slice.start, _slice.stop
        elif len(k) == 1:
            return k[0], None, None
    else:
        return k, None, None


# TODO: alternate indexing specified as indexing function is more open-closed
#   Note that offset has little use since usually key-dependent (here it is fixed)
#   indexing func could include (optinally) key to cover key-dependent offsets
@dataclass
class AudioSegments:
    """A class to access waveform segments through a obj[audio_src, bt:tt] interface
    - flexibility of audio source: specify how to get from src specification to a
        (wf, sr) pair
    - caches: if the same audio_src is asked multiple times in a row, it will keep in
        memory for faster access
    - specify what affine space to use to index segments (default would be samples but
        can specify seconds, utc, etc.)

    ```
    import soundfile as sf
    from hear.tools import AudioSegments

    segs = AudioSegments(src_to_wfsr=sf.read, index_to_seconds_scale=1 / 1e6)
    # Note: 1 / 1e6 to index wf in micro-seconds
    wf = segs[wav_filepath, 1056189.407, 2112556.8]
    # equivalently, use slices
    wf2 = segs[wav_filepath, 1056189.407:2112556.8]
    assert all(wf == wf2)
    ```

    Note: Keys can be [wf_key], [wf_key, bt:tt], or [wf_key, bt, tt].
    Warning: If your wf_key is a tuple, you cannot use [wf_key] since it will be
    confused with the other forms. In this case, do [wf_key,] or [tuple([wf_key])]

    Note: The default index unit is seconds. If you want to index by samples (which
    have less problems due to floating paint arithmetic), you CAN do
    ```
    segs = AudioSegments(..., index_to_seconds_scale=1 / sample_rate)
    ```
    but you should avoid this (unnecessary overhead of still float point arithmetic).
    Instead, you should just retrieve the whole waveform and index it directly:
    ```
    segs = AudioSegments(...)
    segs[wav_filepath][1000:1100]
    ```
    """

    src_to_wfsr: Callable[[AudioSource], Waveform] = sf.read
    index_to_seconds_scale: Number = 1
    index_to_seconds_offset: Number = 0
    _key_to_src_bt_tt: SegmentKeyNormalizer = dflt_key_to_src_bt_tt

    def __getitem__(self, k):
        src, bt, tt = self._key_to_src_bt_tt(k)
        wf, indexer = src_to_wf_and_indexer(
            # note: cached so repeated retrieval of same src will be efficient
            # (already in memory)
            src,
            index_to_seconds_scale=self.index_to_seconds_scale,
            index_to_seconds_offset=self.index_to_seconds_offset,
            src_to_wfsr=self.src_to_wfsr,
        )
        if bt is not None:
            bt = int(indexer(bt))
        if tt is not None:
            tt = int(indexer(tt))
        return wf[bt:tt]


@lru_cache(maxsize=1)
def src_to_wf_and_indexer(
    src: AudioSource, index_to_seconds_scale, index_to_seconds_offset, src_to_wfsr,
):
    wf, sr = src_to_wfsr(src)
    indexer = AffineConverter(
        scale=sr * index_to_seconds_scale, offset=index_to_seconds_offset
    )
    return wf, indexer


def wf_func_to_wfsr_func(wf_func=None, *, sr):
    """Transform a wf returning function so it returns (wf, sr) pairs.

    Sometimes you don't have a function providing (wf, sr) in hand, but instead,
    a function providing wfs (and you know what the sr is).

    >>> wf_func = lambda n: list(range(n))
    >>> wf_func(4)
    [0, 1, 2, 3]

    To use AudioSegment, you'll need to provide a wfsr function though.
    To do so easily from your wf function, we give you wf_func_to_wfsr_func

    >>> wfsr_func = wf_func_to_wfsr_func(wf_func, sr=44100)
    >>> wfsr_func(4)
    ([0, 1, 2, 3], 44100)
    """
    if wf_func is None:
        return partial(wf_func_to_wfsr_func, sr=sr)

    @wraps(wf_func)
    def wfsr_func(*args, **kwargs):
        return wf_func(*args, **kwargs), sr

    return wfsr_func


class AffineConverter(object):
    """
    Getting a callable that will perform an affine conversion.
    Note, it does it as
        (val - offset) * scale
    (Note slope-intercept style (though there is the .from_slope_and_intercept constructor method for that)

    Inverse is available through the inv method, performing:
        val / scale + offset

    >>> convert = AffineConverter(scale=0.5, offset=1)
    >>> convert(0)
    -0.5
    >>> convert(10)
    4.5
    >>> convert.inv(4)
    9.0
    >>> convert.inv(4.5)
    10.0
    """

    def __init__(self, scale=1.0, offset=0.0):
        self.scale = scale
        self.offset = offset

    @classmethod
    def from_slope_and_intercept(cls, slope=1.0, intercept=0.0):
        cls(offset=-intercept / slope, scale=slope)

    def __call__(self, x):
        return (x - self.offset) * self.scale

    def inv(self, x):
        return x / self.scale + self.offset

    def map(self, seq):
        return (self(x) for x in seq)

    def invmap(self, seq):
        return (self.inv(x) for x in seq)



"""This module exists so that users of py2store.ext.audio (now deprecated) can rely on it.

No promises that it will live forever,
so switch to `hear` package as soon as you have the opportunity.
"""
from io import BytesIO
from py2store.stores.local_store import LocalBinaryStore
from py2store.trans import add_wrapper_method

import soundfile as sf

# from py2store.mint import wraps, _empty_func

DFLT_DTYPE = 'int16'
DFLT_FORMAT = 'WAV'
DFLT_N_CHANNELS = 1

# TODO: Do some validation and smart defaults with these
dtype_from_sample_width = {
    1: 'int16',
    2: 'int16',
    3: 'int32',
    4: 'int32',
    8: 'float64',
}

sample_width_for_soundfile_subtype = {
    'DOUBLE': 8,
    'FLOAT': 4,
    'G721_32': 4,
    'PCM_16': 2,
    'PCM_24': 3,
    'PCM_32': 4,
    'PCM_U8': 1,
}

# soundfile_signature not used yet, but intended for a future version of this module, that will use minting
# and signature injection instead of long copy pastes of
soundfile_signature = dict(
    dtype=DFLT_DTYPE, format=DFLT_FORMAT, subtype=None, endian=None
)


class SampleRateAssertionError(ValueError):
    ...


class ReadAudioFileMixin:
    read_kwargs = {}

    def _obj_of_data(self, data):
        return sf.read(BytesIO(data), **self.read_kwargs)


class PcmSerializationTrans:
    def __init__(
        self,
        sr,
        channels=DFLT_N_CHANNELS,
        dtype=DFLT_DTYPE,
        format='RAW',
        subtype='PCM_16',
        endian=None,
    ):
        assert isinstance(sr, int), 'assert_sr must be an int'
        self.sr = sr
        self._rw_kwargs = dict(
            samplerate=sr,
            channels=channels,
            dtype=dtype,
            format=format,
            subtype=subtype,
            endian=endian,
        )

    def _obj_of_data(self, data):
        return sf.read(BytesIO(data), **self._rw_kwargs)[0]


@add_wrapper_method
class WfSrSerializationTrans:
    """An audio serialization/deserialization transformer object whose external objects are (waveform, sample_rate)
    pairs, and internal data some bytes-encoding of the latter.

    See WavSerializationTrans for explanations and doctest examples.
    """

    _read_format = DFLT_FORMAT
    _rw_kwargs = dict(dtype=DFLT_DTYPE, subtype=None, endian=None)

    def __init__(self, dtype=DFLT_DTYPE, format=DFLT_FORMAT, subtype=None, endian=None):
        self._read_format = format
        self._rw_kwargs = dict(dtype=dtype, subtype=subtype, endian=endian)

    def _obj_of_data(self, data):
        return sf.read(BytesIO(data), **self._rw_kwargs)

    def _data_of_obj(self, obj):
        wf, sr = obj
        b = BytesIO()
        sf.write(b, wf, samplerate=sr, format=self._rw_kwargs)
        b.seek(0)
        return b.read()


@add_wrapper_method
class WfSerializationTrans(WfSrSerializationTrans):
    """An audio serializatiobn object like WfSrSerializationTrans, but working with waveforms only (sample rate fixed).

    See WavSerializationTrans for explanations and doctest examples.
    """

    def __init__(
        self,
        assert_sr=None,
        dtype=DFLT_DTYPE,
        format=DFLT_FORMAT,
        subtype=None,
        endian=None,
    ):
        super().__init__(dtype, format, subtype, endian)
        self.assert_sr = assert_sr

    def _obj_of_data(self, data):
        wf, sr = super()._obj_of_data(data)
        if self.assert_sr is not None and sr != self.assert_sr:
            raise SampleRateAssertionError(
                f'{self.assert_sr} expected but I encountered {sr}'
            )
        return wf

    def _data_of_obj(self, obj):
        return super()._data_of_obj((obj, self.sr))


@add_wrapper_method
class WavSerializationTrans:
    r"""A wav serialization/deserialization transformer.

    First let's make a very short waveform.

    >>> from numpy import sin, arange, pi
    >>> n_samples = 5; sr = 44100;
    >>> wf = sin(arange(n_samples) * 2 * pi * 440 / sr)
    >>> wf
    array([0.        , 0.06264832, 0.12505052, 0.18696144, 0.24813785])

    An instance of ``WavSerializationTrans`` will allow you to
    >>> trans = WavSerializationTrans(assert_sr=sr)  # if you want to write data you NEED to specify assert_sr
    >>> wav_bytes = trans._data_of_obj(wf)
    >>> wav_bytes[:44]  # the header bytes
    b'RIFF.\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00D\xac\x00\x00\x88X\x01\x00\x02\x00\x10\x00data\n\x00\x00\x00'
    >>> wav_bytes[44:]  # the data bytes (5 * 2 = 10 bytes)
    b'\x00\x00\x04\x08\x01\x10\xee\x17\xc2\x1f'

    >>> wf_read_from_bytes = trans._obj_of_data(wav_bytes)
    >>> wf_read_from_bytes
    array([   0, 2052, 4097, 6126, 8130], dtype=int16)

    Note that we've serialized floats, but they were deserialized as int16.
    This is the default behavior, but is cusomizable through dtype, subtype, etc.
    With this default dtype=int16 setting though, if you serialize int16 arrays, you'll recover them exactly.

    >>> assert all(trans._obj_of_data(trans._data_of_obj(wf_read_from_bytes)) == wf_read_from_bytes)

    The most common use of WavSerializationTrans through, is to make a class decorator for a store that
    provides wav bytes.

    >>> @WavSerializationTrans.wrapper(assert_sr=sr)
    ... class MyWavStore(dict):
    ...     pass
    >>> my_wav_store = MyWavStore(just_one=wav_bytes)
    >>> my_wav_store['just_one']
    array([   0, 2052, 4097, 6126, 8130], dtype=int16)


    """
    _rw_kwargs = dict(format='WAV', subtype=None, endian=None)
    _read_kwargs = dict(dtype=DFLT_DTYPE)

    def __init__(
        self, assert_sr=None, dtype=DFLT_DTYPE, format='WAV', subtype=None, endian=None,
    ):
        if assert_sr is not None:
            assert isinstance(assert_sr, int), 'assert_sr must be an int'
        self.assert_sr = assert_sr
        self._rw_kwargs = dict(format=format, subtype=subtype, endian=endian)
        self._read_kwargs = dict(dtype=dtype)

    def _obj_of_data(self, data):
        wf, sr = sf.read(BytesIO(data), **self._read_kwargs)
        if self.assert_sr != sr:
            if (
                self.assert_sr is not None
            ):  # Putting None check here because less common, so more efficient on avg
                raise SampleRateAssertionError(
                    f'sr was {sr}, should be {self.assert_sr}'
                )
        return wf

    def _data_of_obj(self, obj):
        wf = obj
        b = BytesIO()
        sf.write(b, wf, samplerate=self.assert_sr, **self._rw_kwargs)
        b.seek(0)
        return b.read()


PcmSerializationMixin = PcmSerializationTrans  # alias for back-compatibility
WfSrSerializationMixin = WfSrSerializationTrans  # alias for back-compatibility
WavSerializationMixin = WavSerializationTrans  # alias for back-compatibility
WfSerializationMixin = WfSerializationTrans  # alias for back-compatibility


class WavLocalFileStore(WavSerializationTrans, LocalBinaryStore):
    def __init__(
        self,
        path_format,
        assert_sr=None,
        max_levels=None,
        dtype=DFLT_DTYPE,
        format='WAV',
        subtype=None,
        endian=None,
    ):
        LocalBinaryStore.__init__(self, path_format, max_levels)
        WavSerializationTrans.__init__(
            self,
            assert_sr=assert_sr,
            dtype=dtype,
            format=format,
            subtype=subtype,
            endian=endian,
        )


WavLocalFileStore2 = WavLocalFileStore  # back-compatibility alias

from py2store import FilesOfZip, filt_iter


def has_wav_extension(string):
    return string.endswith('.wav') and not string.startswith('__MACOSX/')


@filt_iter(filt=has_wav_extension)
class ZipOfWavs(FilesOfZip):
    """Only lists .wav files and provides values as bytes.

    To get numerical waveform, use `WavSerializationTrans.wrapper`

    >>> @WavSerializationTrans.wrapper(assert_sr=44100)
    ... class WfOfZipOfWavs(ZipOfWavs):
    ...     pass
    """

    pass


@WfSrSerializationTrans.wrapper()
class WfSrOfZipOfWavs(ZipOfWavs):
    """A KvReader that provides the (wf, sr) pairs of the .wav files in a zip file"""

    pass


def _length_and_sr_of_wavs(z):
    """A dataframe containing information about the wavfiles of the files in the wfsr store
    Note: Don't depend on this yet -- it's in motion.
    """
    import pandas as pd

    def gen():
        for k, (wf, sr) in z.items():
            yield {'file': k, 'n_samples': len(wf), 'sr': sr}

    df = pd.DataFrame(list(gen()))
    df = df.set_index('file')
    df['duration_s'] = df['n_samples'] / df['sr']
    return df


from dol import mk_dirs_if_missing
import os


@mk_dirs_if_missing
class PcmSourceSessionBlockStore(LocalBinaryStore):
    sep = os.path.sep
    path_depth = 3

    def _id_of_key(self, k):
        raise DeprecationWarning('Deprecated')
        assert len(k) == self.path_depth
        return super()._id_of_key(self.sep.join(self.path_depth * ['{}']).format(*k))

    def _key_of_id(self, _id):
        raise DeprecationWarning('Deprecated')
        key = super()._key_of_id(_id)
        return tuple(key.split(self.sep))



"""
Useful stores (dols; Data Object Layers) for audio storage management.
"""
from io import BytesIO
from py2store.stores.local_store import LocalBinaryStore
from dol.trans import add_wrapper_method
from dol.signatures import Sig, call_forgivingly


import soundfile as sf

from hear.util import (
    DFLT_DTYPE,
    DFLT_FORMAT,
    DFLT_N_CHANNELS,
    SampleRateAssertionError,
    SampleRateMissing,
)


class ReadAudioFileMixin:
    read_kwargs = {}

    def _obj_of_data(self, data):
        return sf.read(BytesIO(data), **self.read_kwargs)


class PcmSerializationTrans:
    def __init__(
        self,
        sr,
        channels=DFLT_N_CHANNELS,
        dtype=DFLT_DTYPE,
        format='RAW',
        subtype='PCM_16',
        endian=None,
    ):
        assert isinstance(sr, int), 'sr must be an int'
        self.sr = sr
        self._rw_kwargs = dict(
            samplerate=sr,
            channels=channels,
            dtype=dtype,
            format=format,
            subtype=subtype,
            endian=endian,
        )

    def _obj_of_data(self, data):
        return sf.read(BytesIO(data), **self._rw_kwargs)[0]


@add_wrapper_method
class WfSrSerializationTrans:
    """An audio serialization/deserialization transformer object whose external objects are (
    waveform, sample_rate)
    pairs, and internal data some bytes-encoding of the latter.

    See WavSerializationTrans for explanations and doctest examples.
    """

    # _read_format = DFLT_FORMAT
    # _rw_kwargs = dict(dtype=DFLT_DTYPE, subtype=None, endian=None)

    def __init__(self, dtype=DFLT_DTYPE, format=DFLT_FORMAT, subtype=None, endian=None):
        self._r_kwargs = dict(dtype=dtype, subtype=subtype, endian=endian)
        self._w_kwargs = dict(format=format, subtype=subtype, endian=endian)

    def _obj_of_data(self, data):
        return sf.read(BytesIO(data), **self._r_kwargs)

    def _data_of_obj(self, obj):
        wf, sr = obj
        b = BytesIO()
        sf.write(b, wf, samplerate=sr, **self._w_kwargs)
        b.seek(0)
        return b.read()


from typing import Optional


@add_wrapper_method
class WfsrToWfWithSrAssertionTrans:
    """Asserts a fixed specified sr (if given) and returns only the wf part of the (wf, sr) data"""

    def __init__(self, assert_sr: Optional[int] = None):
        assert assert_sr is None or isinstance(assert_sr, int), (
            f'assert_sr must be None or an integer. Was ' f'{assert_sr}'
        )
        self.assert_sr = assert_sr

    def wfsr_to_wf_with_sr_assertion(self, wfsr):
        wf, sr = wfsr
        if self.assert_sr != sr:
            if (
                self.assert_sr is not None
            ):  # Putting None check here because less common, so more efficient on avg
                raise SampleRateAssertionError(
                    f'sr was {sr}, should be {self.assert_sr}'
                )
        return wf

    def _obj_of_data(self, data):
        return self.wfsr_to_wf_with_sr_assertion(data)

    def _data_of_obj(self, obj):
        if self.assert_sr is None:
            raise SampleRateMissing(
                f'To write data you need to specify an assert_sr sample rate'
            )
        return obj, self.assert_sr


@add_wrapper_method
class WfSerializationTrans(WfSrSerializationTrans):
    """An audio serialization object like WfSrSerializationTrans, but working with waveforms
    only (sample rate fixed).

    See WavSerializationTrans for explanations and doctest examples.
    """

    def __init__(
        self,
        assert_sr=None,
        dtype=DFLT_DTYPE,
        format=DFLT_FORMAT,
        subtype=None,
        endian=None,
    ):
        super().__init__(dtype, format, subtype, endian)
        self.assert_sr = assert_sr

    def _obj_of_data(self, data):
        wf, sr = super()._obj_of_data(data)
        if self.assert_sr is not None and sr != self.assert_sr:
            raise SampleRateAssertionError(
                f'{self.assert_sr} expected but I encountered {sr}'
            )
        return wf

    def _data_of_obj(self, obj):
        if self.assert_sr is None:
            raise SampleRateMissing(
                f'To write data you need to specify an assert_sr sample rate'
            )
        return super()._data_of_obj((obj, self.assert_sr))


# TODO: Make this with above elements:
#  For example, with @WfsrToWfWithSrAssertionTrans.wrapper(assert_sr=None)
@add_wrapper_method
class WavSerializationTrans(WfSrSerializationTrans, WfsrToWfWithSrAssertionTrans):
    r"""A wav serialization/deserialization transformer.

    First let's make a very short waveform.

    >>> from numpy import sin, arange, pi
    >>> n_samples = 5; sr = 44100;
    >>> wf = sin(arange(n_samples) * 2 * pi * 440 / sr)
    >>> wf
    array([0.        , 0.06264832, 0.12505052, 0.18696144, 0.24813785])

    An instance of ``WavSerializationTrans`` will allow you to

    >>> trans = WavSerializationTrans(assert_sr=sr)

    If you want to write data you NEED to specify assert_sr

    >>> wav_bytes = trans._data_of_obj(wf)
    >>> wav_bytes[:44]  # the header bytes
    b'RIFF.\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00D\xac\x00\x00\x88X\x01\x00\x02\x00\x10\x00data\n\x00\x00\x00'

    >>> wav_bytes[44:]  # the data bytes (5 * 2 = 10 bytes)
    b'\x00\x00\x04\x08\x01\x10\xee\x17\xc2\x1f'

    >>> wf_read_from_bytes = trans._obj_of_data(wav_bytes)
    >>> wf_read_from_bytes
    array([   0, 2052, 4097, 6126, 8130], dtype=int16)

    Note that we've serialized floats, but they were deserialized as int16.
    This is the default behavior, but is cusomizable through dtype, subtype, etc.
    With this default dtype=int16 setting though, if you serialize int16 arrays, you'll recover
    them exactly.

    >>> assert all(trans._obj_of_data(trans._data_of_obj(wf_read_from_bytes)) == wf_read_from_bytes)

    The most common use of WavSerializationTrans through, is to make a class decorator for a
    store that
    provides wav bytes.

    >>> @WavSerializationTrans.wrapper(assert_sr=sr)
    ... class MyWavStore(dict):
    ...     pass
    >>> my_wav_store = MyWavStore(just_one=wav_bytes)
    >>> my_wav_store['just_one']
    array([   0, 2052, 4097, 6126, 8130], dtype=int16)

    """

    def __init__(
        self, assert_sr=None, dtype=DFLT_DTYPE, format='WAV', subtype=None, endian=None,
    ):
        WfsrToWfWithSrAssertionTrans.__init__(self, assert_sr=assert_sr)
        WfSrSerializationTrans.__init__(
            self, dtype=dtype, format=format, subtype=subtype, endian=endian
        )

    def _obj_of_data(self, data):
        wf, sr = super()._obj_of_data(data)
        return WfsrToWfWithSrAssertionTrans._obj_of_data(self, (wf, sr))

    def _data_of_obj(self, obj):
        wf, sr = WfsrToWfWithSrAssertionTrans._data_of_obj(self, obj)
        return super()._data_of_obj((wf, sr))


PcmSerializationMixin = PcmSerializationTrans  # alias for back-compatibility
WfSrSerializationMixin = WfSrSerializationTrans  # alias for back-compatibility
WavSerializationMixin = WavSerializationTrans  # alias for back-compatibility
WfSerializationMixin = WfSerializationTrans  # alias for back-compatibility


class WavLocalFileStore(WavSerializationTrans, LocalBinaryStore):
    def __init__(
        self,
        path_format,
        assert_sr=None,
        max_levels=None,
        dtype=DFLT_DTYPE,
        format='WAV',
        subtype=None,
        endian=None,
    ):
        LocalBinaryStore.__init__(self, path_format, max_levels)
        WavSerializationTrans.__init__(
            self,
            assert_sr=assert_sr,
            dtype=dtype,
            format=format,
            subtype=subtype,
            endian=endian,
        )


WavLocalFileStore2 = WavLocalFileStore  # back-compatibility alias

from py2store.slib.s_zipfile import FilesOfZip, ZipStore
from dol import filt_iter


def has_wav_extension(string):
    return string.endswith('.wav') and not string.startswith('__MACOSX/')


@filt_iter(filt=has_wav_extension)
class ZipOfWavs(FilesOfZip):
    """Only lists .wav files and provides values as bytes.

    To get numerical waveform, use `WavSerializationTrans.wrapper`

    >>> @WavSerializationTrans.wrapper(assert_sr=44100)
    ... class WfOfZipOfWavs(ZipOfWavs):
    ...     pass
    """


@WfSrSerializationTrans.wrapper()
class WfSrOfZipOfWavs(ZipOfWavs):
    """A KvReader that provides the (wf, sr) pairs of the .wav files in a zip file"""


@WfSrSerializationTrans.wrapper()
class WfsrZipStore(ZipStore):
    """Read and write (waveform, sample_rate) pairs in a zip file's files.

    >>> from tempfile import gettempdir
    >>> import os
    >>>
    >>> rootdir = gettempdir()
    >>>
    >>> # preparation
    >>> test_zipfile = os.path.join(rootdir, 'zipstore_test_file.zip')
    >>> if os.path.isfile(test_zipfile):
    ...     os.remove(test_zipfile)
    >>> assert not os.path.isfile(test_zipfile)

    Okay, test_zipfile doesn't exist (but will soon...)

    >>> z = WfsrZipStore(test_zipfile)

    See that the file still doesn't exist (it will only be created when we start writing)

    >>> assert not os.path.isfile(test_zipfile)
    >>> list(z)  # z "is" empty (which makes sense?)
    []
    >>> import numpy as np
    >>> z['foo.wav'] = (np.array([0, 1, -2, 3], np.int16), 44100)
    >>> list(z)
    ['foo.wav']
    >>> z['foo.wav']
    (array([ 0,  1, -2,  3], dtype=int16), 44100)
    """


# TODO: Need to give init access to assert_sr (and more?)
@WfSerializationTrans.wrapper()
class WfZipStore(ZipStore):
    """Read and write waveforms in a zip file's files.

    # >>> from tempfile import gettempdir
    # >>> import os
    # >>>
    # >>> rootdir = gettempdir()
    # >>>
    # >>> # preparation
    # >>> test_zipfile = os.path.join(rootdir, 'zipstore_test_file.zip')
    # >>> if os.path.isfile(test_zipfile):
    # ...     os.remove(test_zipfile)
    # >>> assert not os.path.isfile(test_zipfile)
    #
    # Okay, test_zipfile doesn't exist (but will soon...)
    #
    # >>> z = WfZipStore(test_zipfile, assert_sr=44100)
    #
    # See that the file still doesn't exist (it will only be created when we start writing)
    #
    # >>> assert not os.path.isfile(test_zipfile)
    # >>> list(z)  # z "is" empty (which makes sense?)
    # []
    # >>> import numpy as np
    # >>> z['foo.wav'] = np.array([0, 1, -2, 3], np.int16)
    # >>> list(z)
    # ['foo.wav']
    # >>> z['foo.wav']
    # array([ 0,  1, -2,  3], dtype=int16)
    #
    # Notice that the sample rate was neither written, no read.
    # But if you played `foo.wav` it would reflect the 44100 sample rate you specified when
    # constructing the store.
    #
    # You can also see this sample rate if you use a `Wfsr...` store to read the zip file:
    #
    # >>> zz = WfsrZipStore(test_zipfile)
    # >>> z['foo.wav']
    # (array([ 0,  1, -2,  3], dtype=int16), 44100)
    """

    # @Sig.from_objs(
    #         # lambda assert_sr=None: None,
    #         ZipStore,
    #         WfSerializationTrans)
    # def __init__(self, *args, **kwargs):
    #     # call_forgivingly(super().__init__, self, *args, **kwargs)
    #     call_forgivingly(ZipStore.__init__, self, *args, **kwargs)
    #     call_forgivingly(WfSerializationTrans.__init__, self, *args, **kwargs)


def _length_and_sr_of_wavs(z):
    """A dataframe containing information about the wavfiles of the files in the wfsr store
    Note: Don't depend on this yet -- it's in motion.
    """
    import pandas as pd

    def gen():
        for k, (wf, sr) in z.items():
            yield {'file': k, 'n_samples': len(wf), 'sr': sr}

    df = pd.DataFrame(list(gen()))
    df = df.set_index('file')
    df['duration_s'] = df['n_samples'] / df['sr']
    return df


from py2store.stores.local_store import MakeMissingDirsStoreMixin
import os


class PcmSourceSessionBlockStore(MakeMissingDirsStoreMixin, LocalBinaryStore):
    sep = os.path.sep
    path_depth = 3

    def _id_of_key(self, k):
        raise DeprecationWarning('Deprecated')
        assert len(k) == self.path_depth
        return super()._id_of_key(self.sep.join(self.path_depth * ['{}']).format(*k))

    def _key_of_id(self, _id):
        raise DeprecationWarning('Deprecated')
        key = super()._key_of_id(_id)
        return tuple(key.split(self.sep))



"""
Serialization for regular panel data (audio, regular-tick time-series, etc.).
"""

import numpy as np
import soundfile as sf
from io import BytesIO


class WrongSampleRate(ValueError):
    """ To be raised when the sample rate is not the one that's expected. """

    pass


class WrongSerializationParams(ValueError):
    pass


def mk_reader_and_writer(
    sr: int,
    format='RAW',
    subtype='PCM_16',
    dtype='int16',
    channels: int = 1,
    endian=None,
    always_2d=False,
):
    """Makes a (bijective) pair of numerical arrays serializer and deserializer functions.
    A function returning bijective panel data reader and writer functions with simple interfaces (all parametrizations
    are fixed): `read(source)` and `write(source, data)`.
    The writer and reader are essentially matrix (or array) serializers and deserializers respectively,
    using the same serialization protocols as waveform (into PCM, WAV, etc.)

    Args:
        sr: Sample rate. When the serialization format handles this information (e.g. WAV format)
            the sample rate is actually written (in WAV, in the header bytes), and asserted on reads
            (that is, if you read a WAV file that doesn't have that exact sample rate in it's header, a
            WrongSampleRate error will be raised.
            When the serialization format doesn't (e.g. RAW format (a.k.a. PCM)), it is ignored both on reads and writes
        format: 'RAW', 'WAV' and others (see soundfile.available_formats() for a full list)
        subtype: 'FLOAT', 'PCM_16' and others (see soundfile.available_subtypes() for a full list)
        dtype: 'float64', 'float32', 'int32', 'int16'
        channels: Number of channels (should equal the number of columns of the data matrices that will be
            serialized -- or 1 if the data is flat)
        endian: see soundfile documentation ({'FILE', 'LITTLE', 'BIG', 'CPU'}, sometimes optional)
        always_2d: By default, reading a mono sound file will return a one-dimensional array. With always_2d=True,
            data is always returned as a two-dimensional array, even if the data has only one channel.

    Returns:
        read(k), write(k, v) functions

    >>> n_channels, dtype = 1, 'float64'
    >>> read, write = mk_reader_and_writer(sr=44100, channels=n_channels, subtype='FLOAT', format='RAW', dtype=dtype)
    >>> data = _random_matrix(n_channels=n_channels, dtype=dtype)
    >>> _test_data_write_read(data, writer=write, reader=read)

    >>> n_channels, dtype = 4, 'int16'
    >>> read, write = mk_reader_and_writer(sr=2, channels=n_channels, subtype='PCM_16', format='RAW', dtype=dtype)
    >>> data = _random_matrix(n_channels=n_channels, dtype=dtype)
    >>> _test_data_write_read(data, writer=write, reader=read)
    """
    if not sf.check_format(format, subtype, endian):
        raise WrongSerializationParams(
            f'Not a valid combo: format={format}, subtype={subtype}, endian={endian}'
        )

    subtype = subtype or sf.default_subtype(format)

    if format == 'RAW':

        def read(k):
            wf, _ = sf.read(
                k,
                samplerate=sr,
                channels=channels,
                format=format,
                subtype=subtype,
                dtype=dtype,
                endian=endian,
                always_2d=always_2d,
            )
            return wf

    else:

        def read(k):
            wf, sr_read = sf.read(k, dtype=dtype, always_2d=always_2d)
            if sr != sr_read:
                raise WrongSampleRate(f'Sample rate was {sr_read}: Expected {sr}')
            return wf

    def write(k, v):
        return sf.write(
            k, v, samplerate=sr, format=format, subtype=subtype, endian=endian
        )

    # add some attributes to the functions, for diagnosis purposes
    read.sr = sr
    read.dtype = dtype
    read.format = format
    write.sr = sr
    write.format = format
    write.subtype = subtype

    return read, write


# TODO: Make it pytest compliant (problem with fixture)
def _test_data_write_read(data, writer, reader):
    b = BytesIO()
    writer(b, data)
    b.seek(0)  # rewind
    read_data = reader(b)
    if isinstance(read_data, tuple):
        read_data, read_sr = read_data
    assert np.allclose(read_data, data), 'np.allclose(read_data, data)'
    # assert read_sr == sr, 'read_sr == sr'


def _random_matrix(
    n_samples=100, n_channels=1, value_range=(-2000, 2000), dtype='float64'
):
    """
    Make a random matrix with n_samples rows and n_channels columns, with numbers drawn randomly from
    the value_range interval
    Args:
        n_samples: number of rows in the matrix
        n_channels: number of columns in the matrix (if n_channels=1, the function will output a flat array)
        value_range: the range to pick from

    Returns:
        a matrix (i.e. array of arrays) or a flat array (if n_channels==1).
    """

    if isinstance(value_range, (int, float)):
        interval_length = value_range
        value_range = (-interval_length / 2, interval_length / 2)
    else:
        interval_length = value_range[1] - value_range[0]

    data = (np.random.rand(n_samples, n_channels) * interval_length) + value_range[0]

    if n_channels == 1:
        data = np.ravel(data)

    return data.astype(dtype)


def _random_data_and_serialization_params(
    n_samples=100, n_channels=1, value_range=(-2000, 2000), dtype='float64'
):
    """ Get random data and serialization params (i.e. how to map to bytes)"""
    raise NotImplementedError('Not implemented yet')


if __name__ == '__main__':
    n_channels, dtype = 1, 'float64'
    read, write = mk_reader_and_writer(
        sr=44100, channels=n_channels, subtype='FLOAT', format='RAW', dtype=dtype,
    )
    data = _random_matrix(n_channels=n_channels, dtype=dtype)
    _test_data_write_read(data, writer=write, reader=read)

    n_channels, dtype = 1, 'int16'
    read, write = mk_reader_and_writer(
        sr=44100, channels=n_channels, subtype='PCM_16', format='RAW', dtype=dtype,
    )
    data = _random_matrix(n_channels=n_channels, dtype=dtype)
    _test_data_write_read(data, writer=write, reader=read)

    n_channels, dtype = 1024, 'float32'
    read, write = mk_reader_and_writer(
        sr=10, channels=n_channels, subtype='FLOAT', format='RAW', dtype=dtype
    )
    data = _random_matrix(n_channels=n_channels, dtype=dtype)
    _test_data_write_read(data, writer=write, reader=read)

    n_channels = int(2 ** 10 - 1)  # one more would be too much for format='WAV'
    read, write = mk_reader_and_writer(
        sr=1, channels=n_channels, subtype='FLOAT', format='WAV'
    )
    data = _random_matrix(n_channels=n_channels)
    _test_data_write_read(data, writer=write, reader=read)



"""
Stores that specialize in block storage;
That is, where continuous signal is stored over multiple blocks of data.
"""
from sortedcontainers import SortedList
from collections.abc import Iterable, Mapping
from functools import lru_cache, cached_property
import soundfile as sf
from io import BytesIO
import json
import os
from math import ceil

# from ocore.utils import utime

from py2store.stores.local_store import RelativePathFormatStore
from py2store.mixins import ReadOnlyMixin, IterBasedSizedContainerMixin
from py2store.parse_format import match_re_for_fstring

DFLT_TIME_UNITS_PER_SEC = int(1e6)


class LocalPathStore(ReadOnlyMixin, RelativePathFormatStore):
    """A RelativePathFormatStore, but with write/delete functionality disabled"""

    def head(self):
        return next(iter(self.items()))


def wf_of_block_bytes(block_bytes, sr):
    return sf.read(
        BytesIO(block_bytes),
        dtype='int16',
        channels=1,
        samplerate=sr,
        format='RAW',
        subtype='PCM_16',
    )[0]


def int_if_not_none(x):
    if x is not None:
        return int(x)
    else:
        return None


class SessionBlockStore(LocalPathStore):
    def __init__(self, channel_data_dir, sr=None, **kwargs):
        # figuring out the sample rate
        channel_store = LocalPathStore(channel_data_dir)
        if 'config.json' in channel_store:
            config = json.loads(channel_store['config.json'])
            self.sr = int(config.get('sr', sr))
        else:
            self.sr = sr
        assert (
            self.sr is not None
        ), "I couldn't figure out the sample rate. Should be in a config.json file or given explicitly as argument"

        # Here we make our path inclusion condition more specific, in order to avoid including any files that
        # are not "block" files that are compliant with the folder structure.
        # Namely, we'll only take filepath show structure is CHANNEL_ROOT/s/SESSION/b/BLOCK
        # and further, constrain valid SESSION and BLOCK names to be exactly 16 digits.
        # By tagging the session and block segments in the path_format, we're also enabling the construction of
        # a regular expression that can be used to extract information from our keys.
        # See the get_session_and_block for how it can be used.
        path_format = os.path.join(channel_data_dir, 's/{session:16d}/b/{block:16d}')
        self._path_match_re = match_re_for_fstring(path_format)

        super().__init__(path_format, **kwargs)

    def get_session_and_block(self, k):
        """Extract (session, block) tuple"""
        m = self._path_match_re.search(self._id_of_key(k))
        if m:
            d = m.groupdict()
            return (
                int_if_not_none(d.get('session', None)),
                int_if_not_none(d.get('block', None)),
            )
        else:
            return None, None


class WfStore(SessionBlockStore):
    def __init__(self, channel_data_dir, sr=None):
        super().__init__(channel_data_dir=channel_data_dir, sr=sr, mode='b')

    def __getitem__(self, k):
        block_bytes = super().__getitem__(k)  # the raw bytes
        return wf_of_block_bytes(block_bytes, sr=self.sr)


class TxyzStore(SessionBlockStore):
    @staticmethod
    def df_from_json_lines(data):
        import pandas as pd

        return pd.DataFrame(
            list(map(json.loads, filter(None, data.split('\n'))))
        ).set_index('timestamp')

    def __init__(self, channel_data_dir, sr=None):
        super().__init__(channel_data_dir=channel_data_dir, sr=sr)

    def __getitem__(self, k):
        block_bytes = super().__getitem__(k)  # the raw bytes
        return self.df_from_json_lines(block_bytes)


class DictKeyMap:
    def __init__(self, id_of_key):
        self.__id_of_key = id_of_key.get
        self.__key_of_id = {_id: k for k, _id in id_of_key.items()}.get

    def _id_of_key(self, k):
        return super()._id_of_key(self.__id_of_key(k))

    def _key_of_id(self, _id):
        return self.__key_of_id(super()._key_of_id(_id))


inf = float('infinity')


# TODO: The class below isn't general at all. It assumes, for instance microseconds blocks.
# TODO: Need to factor out the indexing concern (including the assumption of _key_of_block property existence)
class BlocksSearchResult:
    def __init__(
        self, _block_store, bt=-inf, tt=inf, _inclusive=(True, False), _sort_key=None,
    ):
        """
        Object that contains and implements an interval query of a block store.
        :param _block_store: A store. Must be iterable, have a __getitem__, and have sr and time_units_per_sec attrs
        :param bt: the bottom time of the time interval query
        :param tt: the top time of the time interval query
        :param _inclusive: Whether to include or exclude the bt and/or tt
        :param _sort_key: What to sort on.
        >>> from collections import UserDict
        >>> _block_store = UserDict({1000: [1,2,3,4], 1008: [5,6,7,8,9]})
        >>> _block_store.sr = 44100
        >>> _block_store.time_units_per_sec = 88200
        >>>
        >>> r = BlocksSearchResult(_block_store, bt=1004, tt=1014)
        >>> list(r.values())
        [[3, 4], [5, 6, 7]]
        >>> assert list(BlocksSearchResult(_block_store).values()) == [[1, 2, 3, 4], [5, 6, 7, 8, 9]]
        >>> assert list(BlocksSearchResult(_block_store, bt=1002, tt=20000).values()) == [[2, 3, 4], [5, 6, 7, 8, 9]]
        >>> assert list(BlocksSearchResult(_block_store, bt=900, tt=1000).values()) == []
        >>> assert list(BlocksSearchResult(_block_store, bt=900, tt=1004).values()) == [[1, 2]]
        >>> assert list(BlocksSearchResult(_block_store, bt=1004, tt=1008).values()) == [[3, 4]]
        >>> assert list(BlocksSearchResult(_block_store, bt=1004, tt=1009).values()) == [[3, 4], [5]]
        >>> assert list(BlocksSearchResult(_block_store, bt=1004, tt=1010).values()) == [[3, 4], [5]]
        >>> assert list(BlocksSearchResult(_block_store, bt=1004, tt=1011).values()) == [[3, 4], [5, 6]]

        """
        self._block_store = _block_store
        self.bt = bt
        self.tt = tt
        self.sr = self._block_store.sr  # sample rate of the block data
        self.time_units_per_sec = (
            self._block_store.time_units_per_sec
        )  # sample rate of the query time unit
        self._inclusive = _inclusive
        # _sort_key is not used here, but can be used to tell SortedList how to sort keys outputed by __iter__
        self._sort_key = _sort_key

    @cached_property
    def _key_of_block(self):
        if hasattr(self._block_store, '_key_of_block'):
            return self._block_store._key_of_block
        else:
            return SortedList(self._block_store.__iter__(), self._sort_key)

    @cached_property
    def intersecting_blocks(self):
        """ Iterator of blocks that CONTAIN the [bt, tt) range """
        # TODO: Figure out the most efficient way to do this
        bottom_idx = self._key_of_block.bisect_right(self.bt)
        if bottom_idx > 0:
            bottom_idx -= 1
        min_block = self._key_of_block[bottom_idx]
        return list(
            self._key_of_block.irange(
                minimum=min_block, maximum=self.tt, inclusive=self._inclusive
            )
        )

    def __iter__(
        self,
    ):  # TODO: Change to be the bts of the values returned by items, not block bts.
        return self.items()

    def blocks_intersecting_bt_tt(self):
        """ Iterator of blocks that CONTAIN the [bt, tt) range """
        # TODO: Figure out the most efficient way to do this
        bottom_idx = self._key_of_block.bisect_right(self.bt)
        if bottom_idx > 0:
            bottom_idx -= 1
        min_block = self._key_of_block[bottom_idx]
        return self._key_of_block.irange(
            minimum=min_block, maximum=self.tt, inclusive=self._inclusive
        )

    def items(self):
        samples_per_ts_unit = self.sr / self.time_units_per_sec
        assert self.bt < self.tt, 'you entered a bt >= tt'

        ts_gen = iter(self.intersecting_blocks)
        # getting the first block
        try:
            current_block_ts = next(ts_gen)
            while current_block_ts < self.tt:
                wf = self._block_store[current_block_ts]
                wf_len_ts_unit = len(wf) / samples_per_ts_unit
                # how much to remove from the left and right from the wf in the unit of the block timestamps
                remove_left_ts_unit = max(0, self.bt - current_block_ts)
                remove_right_ts_unit = max(
                    0, current_block_ts + wf_len_ts_unit - self.tt
                )
                # converting into number of samples:
                # is int best here for rounding? Does not matter in our case at hand since no decimal
                remove_left_sample_unit = int(remove_left_ts_unit * samples_per_ts_unit)
                remove_right_sample_unit = int(
                    remove_right_ts_unit * samples_per_ts_unit
                )

                # sadly array[:-0] is empty (of course) instead of everything so I have to take that case apart
                # there must be a better way but running out of time
                if remove_right_ts_unit:
                    wf = wf[remove_left_sample_unit:-remove_right_sample_unit]
                else:
                    wf = wf[remove_left_sample_unit:]
                if len(wf) > 0:
                    wf_bt = current_block_ts + remove_left_ts_unit
                    wf_tt = current_block_ts + wf_len_ts_unit - remove_right_ts_unit
                    yield (wf_bt, wf_tt), wf
                current_block_ts = next(ts_gen)
        except StopIteration:
            pass

    def keys(self):
        for k, _ in self.items():
            yield k

    def values(self):
        for _, v in self.items():
            yield v

    @lru_cache(maxsize=1)
    def __len__(self):
        return super().__len__()

    @lru_cache(maxsize=1)
    def __contains__(self, k):
        return super().__contains__(k)


class BlockWfStore(DictKeyMap, WfStore):
    def __init__(
        self,
        channel_data_dir,
        sr=None,
        time_units_per_sec=DFLT_TIME_UNITS_PER_SEC,
        _sort_key=None,
    ):
        # TODO: Here I construct a WfStore twice. There must be a better way!
        s = WfStore(channel_data_dir, sr)
        key_of_block = {s.get_session_and_block(key)[1]: key for key in s.__iter__()}
        # assert list(key_of_block.keys()) == sorted(key_of_block.keys()), "the blocks are not sorted"
        WfStore.__init__(self, channel_data_dir, sr)
        DictKeyMap.__init__(self, id_of_key=key_of_block)
        self.time_units_per_sec = time_units_per_sec

        self._sort_key = _sort_key

    @cached_property
    def _key_of_block(self):
        return SortedList(self.__iter__(), self._sort_key)

    def block_search(self, bt=-inf, tt=inf):
        return BlocksSearchResult(
            _block_store=self,
            bt=bt,
            tt=tt,
            _inclusive=(True, False),
            _sort_key=self._sort_key,
        )


class SessionStore(LocalPathStore):
    def __init__(
        self,
        session_dir,
        time_units_per_sec=DFLT_TIME_UNITS_PER_SEC,
        csv_timestamp_time_units_per_sec=int(1e3),
        rel_path_format='{session:13d}/t/tags{csv_timestamp:13d}.csv',
        sr=None,
        **kwargs
    ):
        """
        A store for sessions folders.
        :param session_dir: The sessions directory, which contains sessions subfolders.
        :param sr:
        :param time_units_per_sec: The timestamp unit to use for bt and tt, in num of units per second.
            For example, if milliseconds, time_units_per_sec=1000, if microseconds, time_units_per_sec=1000000.
        :param csv_timestamp_time_units_per_sec: "time sample rate" The timestamp unit the csv filename uses.
        :param rel_path_format: The pathformat under the sessions_dir. Is used to filter the files as
            well as match information encoded in filename (such as csv_timestamp)
        :param kwargs: Passed on to the super class __init__
        """
        path_format = os.path.join(session_dir, rel_path_format)

        super().__init__(path_format, **kwargs)

        self._path_match_re = match_re_for_fstring(path_format)
        self.sr = sr
        self.time_units_per_sec = time_units_per_sec
        self.csv_timestamp_time_units_per_sec = csv_timestamp_time_units_per_sec

    def key_info(self, k):
        """Extract dict of info from parsing file"""
        m = self._path_match_re.search(self._id_of_key(k))
        if m:
            return {field: int(val) for field, val in m.groupdict().items()}
        else:
            return {}


class ScoopAnnotationsStore(SessionStore):
    @staticmethod
    def df_of_csv(csv_str):
        from io import StringIO
        import pandas as pd

        return pd.read_csv(StringIO(csv_str))

    def __getitem__(self, k):
        v = super().__getitem__(k)
        df = self.df_of_csv(v)
        csv_timestamp = self.key_info(k).get('csv_timestamp', None)
        if csv_timestamp is None:
            raise ValueError("Couldn't parse out the csv's timestamp.")
        csv_timestamp = int(
            csv_timestamp
            * self.time_units_per_sec
            / self.csv_timestamp_time_units_per_sec
        )
        df['bt'] = csv_timestamp + df['bt'] * self.time_units_per_sec
        df['tt'] = csv_timestamp + df['tt'] * self.time_units_per_sec
        df = df.sort_values(['bt', 'tt'])
        return df


class ScoopAnnotations:
    def __init__(self, annotations_dir):
        import pandas as pd

        store = ScoopAnnotationsStore(annotations_dir)
        self._store = store
        self.annots_df = pd.concat(list(store.values())).sort_values(['bt', 'tt'])
        self.annots_df = self.annots_df.reset_index(drop=True)

    def __getitem__(self, k):
        if not isinstance(k, slice):
            k = slice(k)
        df = self.annots_df
        return df[(df['bt'] >= k.start) & (df['tt'] < k.stop)].to_dict(orient='records')


# if __name__ == '__main__':
#     from otolite.stores import BlocksSearchResult
#     from collections import UserDict
#
#     _block_store = UserDict({1000: [1, 2, 3, 4], 1008: [5, 6, 7, 8, 9]})
#     _block_store.sr = 44100
#     _block_store.time_units_per_sec = 88200
#
#     r = BlocksSearchResult(_block_store, bt=1001, tt=1014)
#     print(list(r.items()))
#
#     assert list(BlocksSearchResult(_block_store, bt=900, tt=1009).values()) == [[1, 2, 3, 4]]



"""
hear utilities
"""
from typing import Sequence, Union, Any, Iterable, Callable
import numpy as np
from numbers import Number

Sample = Number
Waveform = Sequence[Sample]

DFLT_DTYPE = 'int16'
DFLT_FORMAT = 'WAV'
DFLT_N_CHANNELS = 1

# TODO: Do some validation and smart defaults with these
dtype_from_sample_width = {
    1: 'int16',
    2: 'int16',
    3: 'int32',
    4: 'int32',
    8: 'float64',
}

sample_width_for_soundfile_subtype = {
    'DOUBLE': 8,
    'FLOAT': 4,
    'G721_32': 4,
    'PCM_16': 2,
    'PCM_24': 3,
    'PCM_32': 4,
    'PCM_U8': 1,
}

dflt_subtype_for_sample_width = {2: 'PCM_16', 3: 'PCM_24', 4: 'PCM_32', 8: 'DOUBLE'}

# soundfile_signature not used yet, but intended for a future version of this module, that will use minting
# and signature injection instead of long copy pastes of
soundfile_signature = dict(
    dtype=DFLT_DTYPE, format=DFLT_FORMAT, subtype=None, endian=None
)


class SampleRateError(ValueError):
    """Error raised when something is wrong regarding the sample rate (sr)."""


class SampleRateAssertionError(SampleRateError):
    """Error raised when a sample rate's value is not the one that is expected/enforced."""


class SampleRateMissing(SampleRateError):
    """Error raised when a sample rate is missing (and might be required)."""


num_type_synonyms = [
    {
        'dtype': 'int16',
        'soundfile': 'PCM_16',
        'pyaudio': 'paInt16',
        'n_bits': 16,
        'n_bytes': 2,
        'numpy': np.int16,
        'struct': 'h',
    },
    {
        'dtype': 'int8',
        'soundfile': 'PCM_S8',
        'pyaudio': 'paInt8',
        'n_bits': 8,
        'n_bytes': 1,
        'numpy': np.int8,
        'struct': 'b',
    },
    {
        'dtype': 'int24',
        'soundfile': 'PCM_24',
        'pyaudio': 'paInt24',
        'n_bits': 24,
        'n_bytes': 3,
        'numpy': None,
        'struct': None,
    },
    {
        'dtype': 'int32',
        'soundfile': 'PCM_32',
        'pyaudio': 'paInt32',
        'n_bits': 32,
        'n_bytes': 4,
        'numpy': np.int32,
        'struct': 'i',
    },
    {
        'dtype': 'uint8',
        'soundfile': 'PCM_U8',
        'pyaudio': 'paUInt8',
        'n_bits': 8,
        'n_bytes': 1,
        'numpy': np.uint8,
        'struct': 'B',
    },
    {
        'dtype': 'float32',
        'soundfile': 'FLOAT',
        'pyaudio': 'paFloat32',
        'n_bits': 32,
        'n_bytes': 4,
        'numpy': np.float32,
        'struct': 'f',
    },
    {
        'dtype': 'float64',
        'soundfile': 'DOUBLE',
        'pyaudio': None,
        'n_bits': 64,
        'n_bytes': 8,
        'numpy': np.float64,
        'struct': 'd',
    },
]


def num_type_for(num, num_sys='n_bits', target_num_sys='soundfile'):
    """Translate from one (sample width) number type to another.

    :param num:
    :param num_sys:
    :param target_num_sys:
    :return:

    >>> num_type_for(16, "n_bits", "soundfile")
    'PCM_16'
    >>> num_type_for(np.array([1.0, 2.0, 3.0]).dtype, "numpy", "soundfile")
    'PCM_24'

    Tip: Use with `functools.partial` when you have some fix translation endpoints.

    >>> from functools import partial
    >>> get_dtype_from_n_bytes = partial(
    ...     num_type_for, num_sys="n_bytes", target_num_sys="dtype"
    ... )
    >>> get_dtype_from_n_bytes(8)
    'float64'
    """
    for d in num_type_synonyms:
        if num == d[num_sys]:
            if target_num_sys in d:
                return d[target_num_sys]
            else:
                raise ValueError(
                    f'Did not find any {target_num_sys} entry for {num_sys}={num}'
                )
    raise ValueError(f'Did not find any entry for {num_sys}={num}')
